package de.be4.smvclipse.prolog;

import java.io.OutputStream;
import java.io.PrintWriter;
import java.math.BigInteger;
import java.util.Arrays;

/**
 * Helper class to generate Prolog terms.
 * 
 * @author Daniel Plagge
 */
public final class PrologTermOutput {
	private final static char[] VALIDCHARS = validChars();

	private final PrintWriter out;

	// comma_needed states if the next term can be printed directly (false) or
	// if a separating comma is needed first
	private boolean comma_needed = false;

	private final boolean use_indention;
	private int indent_level = 0;
	private int ignore_indention_level = 0;

	// flag to enable printing of terms without arguments as atoms.
	// if set, the last printed object was a functor, and if anything is printed
	// before closing the term, an opening parenthesis should be printed.
	private boolean lazy_parenthesis = false;

	public PrologTermOutput(final PrintWriter out, final boolean use_indention) {
		this.use_indention = use_indention;
		this.out = out;
	}

	public PrologTermOutput(final PrintWriter out) {
		this(out, true);
	}

	public PrologTermOutput(final OutputStream out, final boolean use_indention) {
		this(new PrintWriter(out), use_indention);
	}

	public PrologTermOutput(final OutputStream out) {
		this(new PrintWriter(out));
	}

	/**
	 * Escapes existing apostrophes by backslashes.
	 * 
	 * @param input
	 *            A string, never <code>null</code>.
	 * @param single_quotes
	 *            if single quotes may be used
	 * @param double_quotes
	 *            if double quotes may be used
	 * @return the escaped input string
	 */
	private void escape(String input, boolean single_quotes,
			boolean double_quotes) {
		for (int i = 0; i < input.length(); i++) {
			final char c = input.charAt(i);
			if (Arrays.binarySearch(VALIDCHARS, c) >= 0) {
				out.print(c);
			} else if (c == '\'') {
				out.print(single_quotes ? "'" : "\\'");
			} else if (c == '\\') {
				out.print("\\\\");
			} else if (c == '\n') {
				out.print("\\n");
			} else if (c == '"') {
				out.print(double_quotes ? "\"" : "\\\"");
			} else {
				out.print('\\');
				out.print(Integer.toOctalString(c));
				out.print('\\');
			}
		}
	}

	private static char[] validChars() {
		StringBuilder buf = new StringBuilder();
		buf.append("abcdefghijklmnopqrstuvwxyz");
		buf.append("ABCDEFGHIJKLMNOPQRSTUVWXYZ");
		buf.append("0123456789");
		buf.append("_ +-*/^<>=~:.?@#$&!;%(),[]{|}");
		char[] chars = buf.toString().toCharArray();
		Arrays.sort(chars);
		return chars;
	}

	private static boolean escapeIsNeeded(String input) {
		final int length = input.length();
		if (length > 0 && Character.isLowerCase(input.charAt(0))) {
			for (int i = 1; i < length; i++) {
				char c = input.charAt(i);
				if (c != '_' && !Character.isLetterOrDigit(c)) {
					return true;
				}
			}
			return false;
		} else {
			return true;
		}
	}

	/**
	 * Start a new term. This methods prints the (escaped, if needed) functor
	 * and the opening brackets. All other terms (atoms, numbers, variables,
	 * ...) are printed as arguments of this term until it is closed with
	 * {@link #closeTerm()}.
	 * 
	 * Same as {@link #openTerm(String, boolean)} with <code>false</code> as
	 * second argument.
	 * 
	 * @param functor
	 *            the functor, never <code>null</code>.
	 */
	public void openTerm(String functor) {
		openTerm(functor, false);
	}

	/**
	 * Start a new term. This methods prints the (escaped, if needed) functor
	 * and the opening brackets. All other terms (atoms, numbers, variables,
	 * ...) are printed as arguments of this term until it is closed with
	 * {@link #closeTerm()}. You should close all opened terms.
	 * 
	 * Use this method instead of {@link #openTerm(String)} if you want to
	 * control whether the arguments of the term should be indented or not. This
	 * is useful to write terms more compact when you know that they are always
	 * short.
	 * 
	 * @param functor
	 *            the functor, never <code>null</code>
	 * @param ignoreIndention
	 *            if this is set to true, the arguments of this term are not
	 *            subject to indent.
	 */
	public void openTerm(String functor, boolean ignoreIndention) {
		printAtom(functor);
		lazy_parenthesis = true;
		comma_needed = false;
		indent_level += 2;
		if (ignore_indention_level > 0) {
			ignore_indention_level++;
		} else if (ignoreIndention) {
			ignore_indention_level = 1;
		}
	}

	private void printIndention() {
		if (use_indention && ignore_indention_level == 0) {
			// synchronized to speed up printing
			synchronized (out) {
				out.println();
				for (int i = 0; i < indent_level; i++) {
					out.print(' ');
				}
			}
		}
	}

	/**
	 * Finish a term that was started with {@link #openTerm(String)}. This
	 * method basically prints the closing parenthesis.
	 */
	public void closeTerm() {
		if (lazy_parenthesis) {
			lazy_parenthesis = false;
		} else {
			out.print(')');
		}
		comma_needed = true;
		indent_level -= 2;
		if (ignore_indention_level > 0) {
			ignore_indention_level--;
		}
	}

	/**
	 * Print an atom. The atom will be escaped, if needed.
	 * 
	 * @param content
	 *            the name of the atom, never <code>null</code>
	 */
	public void printAtom(String content) {
		synchronized (out) {
			printCommaIfNeeded();
			if (escapeIsNeeded(content)) {
				out.print('\'');
				escape(content, false, true);
				out.print('\'');
			} else {
				out.print(content);
			}
			comma_needed = true;
		}
	}

	/**
	 * Print a string. The content of the string will be escaped, if needed.
	 * 
	 * @param content
	 *            the string content, never <code>null</code>
	 */
	public void printString(String content) {
		synchronized (out) {
			printCommaIfNeeded();
			out.print('"');
			escape(content, true, false);
			out.print('"');
			comma_needed = true;
		}
	}

	/**
	 * Print a number.
	 * 
	 * @param number
	 *            the number to print
	 */
	public void printNumber(long number) {
		synchronized (out) {
			printCommaIfNeeded();
			out.print(number);
			comma_needed = true;
		}
	}
	/**
	 * Print a number.
	 * 
	 * @param number
	 *            the number to print
	 */
	public void printNumber(BigInteger number) {
		synchronized (out) {
			printCommaIfNeeded();
			out.print(number);
			comma_needed = true;
		}
	}

	/**
	 * Start a new list. All following terms (atoms, numbers, etc.) until the
	 * next call of {@link #closeList()} are put into the list. All opened lists
	 * should be closed. Basically this method prints the opening bracket.
	 */
	public void openList() {
		synchronized (out) {
			printCommaIfNeeded();
			out.print('[');
			comma_needed = false;
			indent_level += 1;
		}
	}

	/**
	 * Finish a list that was started with {@link #openList()}. Basically this
	 * method prints the closing bracket.
	 */
	public void closeList() {
		synchronized (out) {
			out.print(']');
			comma_needed = true;
			indent_level -= 1;
		}
	}

	/**
	 * Print a Prolog variable. Variables should start with an upper case
	 * character (or underscore) and should not contain spaces (and other
	 * illegal characters). For variables, no escaping is done.
	 * 
	 * @param var
	 *            the name of the variable, never <code>null</code>
	 * @throws IllegalArgumentException
	 *             if the variable is not a syntactically valid Prolog variable.
	 */
	public void printVariable(String var) {
		printCommaIfNeeded();
		checkVariable(var);
		out.print(var);
		comma_needed = true;
	}

	private void checkVariable(String var) {
		boolean ok = var.length() > 0;
		if (ok) {
			char c = var.charAt(0);
			ok = c == '_' || Character.isUpperCase(c);
			for (int i = 1; ok && i < var.length(); i++) {
				c = var.charAt(i);
				ok &= c == '_' || Character.isLetterOrDigit(c);
			}
		}
		if (!ok) {
			throw new IllegalArgumentException(
					"Invalid name for Prolog variable '" + var + "'");
		}
	}

	/**
	 * flush the underlying output stream
	 */
	public void flush() {
		out.flush();
	}

	private void printCommaIfNeeded() {
		if (lazy_parenthesis) {
			out.print('(');
			lazy_parenthesis = false;
		}
		if (comma_needed) {
			out.print(", ");
			printIndention();
		}
	}

	/**
	 * print a Prolog full stop.
	 */
	public void fullstop() {
		out.print('.');
		if (use_indention) {
			out.println();
		}
		comma_needed = false;
	}

}
